#!/usr/bin/env lua
-- DESCR: use "CM !varname = value" statements for configuration of fitness function

-- variables tweaking optimization
local vars = {
  z0 = 50.0,              -- cable impedance
  forward_angle = 0.0,    -- 0 = gain in Y-axis only; otherwise use worst gain within given angle (to Y-axis)
  backward_angle = 90.0,  -- 0 = backward gain on negative Y-axis only; otherwise use best gain within angle
  circular = false,       -- "left" = left-handed circular, "right" = right-handed circular polarization
  backward_penalty = 1.0,       -- penalty for backward gain (around negative Y-axis)
  swr_penalty = 2.0,            -- penalty weight for impedance mismatch
  commonmode_penalty = 0.0,     -- penalty weight for common-mode (high impedance feed)
  commonmode_impedance = 45.0,  -- assumed worst-case (low) impedance for common-mode on feed-line
}

-- attenuation in dB due to complex impedance mismatch
-- (assuming 100% loss of reflected energy)
local function swr_penalty(real, imag)
  local r = real / vars.z0
  local i = imag / vars.z0
  return math.log(4*r / (r*r + 2*r + i*i + 1), 10) * 10
end

-- penalty in dB for potential common-mode on feed-line
-- (lower feed impedance results in lower penalty)
local commonmode_penalty
do
  local function calc(real, imag)
    return (
      10.0 * math.log(
        0.25 * (
          (
            1.0 + (
              vars.commonmode_impedance /
              (vars.commonmode_impedance + 0.5 * (real*real + imag*imag)^0.5)
            )
          ) ^ 2.0
        ),
        10.0
      )
    )
  end
  commonmode_penalty = function(...) return calc(...) - calc(vars.z0, 0) end
end

-- worse-case attenutation in dB for linear polarized waves
-- when optimizing for circular polarization
-- (returns nil on wrong orientation, e.g. left-handed instead of right-handed)
local function circular_penalty(theta_abs, theta_arg, phi_abs, phi_arg)
  local stokes = (
    2 * theta_abs * phi_abs * math.sin((theta_arg - phi_arg)/180*math.pi) /
    (theta_abs^2 + phi_abs^2)
  )
  if vars.circular == "left" then stokes = -stokes
  elseif vars.circular ~= "right" then
    error('Fitness variable !circular must be false, "left", or "right"')
  end
  if stokes <= 0 then return nil end
  return 10.0 * math.log(
    1 - math.sqrt(1 - stokes^2),  -- consider worst-case for linear pol.
    10
  )
end

-- variables to store worst-case penalties/attenuations
local worst_swr_penalty = 1.0/0.0
local worst_commonmode_penalty = 1.0/0.0
local worst_forward = 1.0/0.0
local worst_backward = -1.0/0.0

-- helper variables for parser
local state = nil
local done

-- process *.out file of NEC2
for line in io.lines() do
  if line:find(" COMMENTS ") then
    state = "comments"
  elseif line:find(" ANTENNA INPUT PARAMETERS ") then
    state = "impedance"
    done = false
  elseif line:find(" RADIATION PATTERNS ") then
    state = "gain"
    done = false
  elseif state == "comments" then
    if line:find("^ *%-%-%-.*%-%-%- *$") then
      state = nil
    else
      local varname, value = line:match("^ *! *([^= ]*) *= *(.*)$")
      if varname then
        assert(vars[varname] ~= nil, "Unexpected variable name for fitness configuration")
        vars[varname] = assert(load("return " .. value, nil, "t", {}))()
      end
    end
  elseif state == "impedance" then
    local real, imag = line:match("^ +[-+0-9.Ee]+ +[-+0-9.Ee]+ +[-+0-9.Ee]+ +[-+0-9.Ee]+ +[-+0-9.Ee]+ +[-+0-9.Ee]+ +([-+0-9.Ee]+) +([-+0-9.Ee]+) ")
    if real then
      real, imag = tonumber(real), tonumber(imag)
      if real <= 0 then
        print("-99999")
        return
      end
      worst_swr_penalty = math.min(
        worst_swr_penalty, swr_penalty(real, imag)
      )
      worst_commonmode_penalty = math.min(
        worst_commonmode_penalty,
        commonmode_penalty(real, imag)
      )
      done = true
    elseif done then
      state = nil
    end
  elseif state == "gain" then
    local theta, phi, gain, theta_abs, theta_arg, phi_abs, phi_arg = line:match("^ +([-+0-9.Ee]+) +([-+0-9.Ee]+) +[-+0-9.Ee]+ +[-+0-9.Ee]+ +([-+0-9.Ee]+) +[-+0-9.Ee]+ +[-+0-9.Ee]+ +[A-Z]+ +([-+0-9.Ee]+) +([-+0-9.Ee]+) +([-+0-9.Ee]+) +([-+0-9.Ee]+)")
    if gain then
      theta = tonumber(theta)
      phi   = tonumber(phi)
      gain  = tonumber(gain)
      local alpha = math.asin(math.sin(theta/180*math.pi) * math.sin(phi/180*math.pi)) / math.pi * 180
      if alpha >= 89.999 - vars.forward_angle then
        local cp = 0.0
        if vars.circular then
          cp = circular_penalty(
            tonumber(theta_abs),
            tonumber(theta_arg),
            tonumber(phi_abs),
            tonumber(phi_arg)
          )
          if cp == nil then
            print("-99999")
            return
          end
        end
        worst_forward = math.min(worst_forward, gain + cp)
      elseif alpha <= -89.999 + vars.backward_angle then
        worst_backward = math.max(worst_backward, gain)
      end
      done = true
    elseif done then
      state = nil
    end
  end
end

-- calculate penalty for backward radiation
local backward_penalty = math.log(
  10^(worst_forward/20) / ( 10^(worst_forward/20) + 10^(worst_backward/20) ),
  10
) * 20

-- calculate and write total fitness value (using vars.*_penalty as weights)
print(
  worst_forward +
  vars.backward_penalty   * backward_penalty +
  vars.swr_penalty        * worst_swr_penalty +
  vars.commonmode_penalty * worst_commonmode_penalty
)
